//*****************************************************************************
//
// Compile with cl -Ox cpuid.c
//
//*****************************************************************************

#include <windows.h>

#include <stdio.h>
#include <stdlib.h>

//*****************************************************************************

#define  EMU

//*****************************************************************************

void cpuid(DWORD index, DWORD *values)
{
  __asm
  {
    push  esi
    push  ebx

    mov eax,index
    cpuid

    mov esi,values

    mov [esi+0],eax
    mov [esi+4],ebx
    mov [esi+8],ecx
    mov [esi+12],edx

    pop ebx
    pop esi
  }
}

//*****************************************************************************

char *dumpascii(DWORD *values)
{
  static char buffer[20];
  unsigned char *p = (unsigned char *)values;
  char *b = buffer;

  int i;

  for(i=0; i<16; i++)
  {
    if ((*p >= 0x20) && (*p <= 0x7E))
      *b++ = *p;
    else
      *b++ = '.';

    p++;
  }

  *b = 0;

  return(buffer);
}

//*****************************************************************************

int main(int argc, char **argv)
{
  DWORD values[4];
  DWORD i, j;
#ifdef EMU
  FILE *f;

  f = fopen("cpuidemu.asm","w");

  if (f)
  {
    fprintf(f,"\t.386\n");
    fprintf(f,"\n");
    fprintf(f,"\t.MODEL FLAT,C\n");
    fprintf(f,"\n");
    fprintf(f,"\t.DATA\n");
    fprintf(f,"\n");
    fprintf(f,"EmuTable\tequ\t$\n");
  }
#endif

  puts("index      eax      ebx      ecx      edx");

  i = 0x00000000;

  cpuid(i, values);

  printf("%08X - %08X %08X %08X %08X - %s\n", i, values[0],values[1],values[2],values[3], dumpascii(values));

  j = values[0];

  if ((j & 0xFFFFF000) == i) // Sanity check, and limit output to 1023 lines
  {
#ifdef EMU
    fprintf(f,"\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
#endif

    for(i++; i<=j; i++)
    {
      cpuid(i, values);

      printf("%08X - %08X %08X %08X %08X - %s\n", i, values[0],values[1],values[2],values[3], dumpascii(values));

#ifdef EMU
      fprintf(f,"\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
#endif
    }
  }

  putchar('\n');


  i = 0x80000000;

  cpuid(i, values);

  printf("%08X - %08X %08X %08X %08X - %s\n", i, values[0],values[1],values[2],values[3], dumpascii(values));

  j = values[0];

  if ((j & 0xFFFFF000) == i) // Sanity check
  {
#ifdef EMU
    fprintf(f,"\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
#endif

    for(i++; i<=j; i++)
    {
      cpuid(i, values);

      printf("%08X - %08X %08X %08X %08X - %s\n", i, values[0],values[1],values[2],values[3], dumpascii(values));

#ifdef EMU
      fprintf(f,"\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
#endif
    }
  }

  putchar('\n');

#ifdef EMU
  if (f)
  {
    fprintf(f,"\tdd\t0FFFFFFFFh\n");

    fprintf(f,"\n");
    fprintf(f,"\t.CODE\n");
    fprintf(f,"\n");
    fprintf(f,"CPUIDEmu PROC ; On Entry EAX = Index\n");
    fprintf(f,"\n");
    fprintf(f,"        push    esi\n");
    fprintf(f,"        mov     esi,EmuTable\n");
    fprintf(f,"\n");
    fprintf(f,"        cmp     eax,0FFFFFFFFh\n");
    fprintf(f,"        jz      cpuid_30\n");
    fprintf(f,"\n");
    fprintf(f,"cpuid_10:\n");
    fprintf(f,"\n");
    fprintf(f,"        cmp     [esi],eax\n");
    fprintf(f,"        jz      cpuid_20\n");
    fprintf(f,"\n");
    fprintf(f,"        add     esi,20\n");
    fprintf(f,"\n");
    fprintf(f,"        cmp     dword ptr [esi],0FFFFFFFFh\n");
    fprintf(f,"        jnz     cpuid_10\n");
    fprintf(f,"\n");
    fprintf(f,"        jmp     cpuid_30\n");
    fprintf(f,"\n");
    fprintf(f,"cpuid_20:\n");
    fprintf(f,"\n");
    fprintf(f,"        mov     eax,[esi +  4]\n");
    fprintf(f,"        mov     ebx,[esi +  8]\n");
    fprintf(f,"        mov     ecx,[esi + 12]\n");
    fprintf(f,"        mov     edx,[esi + 16]\n");
    fprintf(f,"\n");
    fprintf(f,"cpuid_30:\n");
    fprintf(f,"\n");
    fprintf(f,"        pop     esi\n");
    fprintf(f,"        ret\n");
    fprintf(f,"\n");
    fprintf(f,"CPUIDEmu ENDP\n");
    fprintf(f,"\n");
    fprintf(f,"\tEND\n");

    fclose(f);
  }
#endif

#if 1
  puts("Press any key to continue..");
  getch();
#endif

#if 1
  printf("EmuTable\tequ\t$\n");

  i = 0x00000000;

  cpuid(i, values);

  j = values[0];

  if ((j & 0xFFFFF000) == i) // Sanity check, and limit output to 1023 lines
  {
    printf("\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);

    for(i++; i<=j; i++)
    {
      cpuid(i, values);

      printf("\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
    }
  }

  i = 0x80000000;

  cpuid(i, values);

  j = values[0];

  if ((j & 0xFFFFF000) == i) // Sanity check
  {
    printf("\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);

    for(i++; i<=j; i++)
    {
      cpuid(i, values);

      printf("\tdd\t0%08Xh,0%08Xh,0%08Xh,0%08Xh,0%08Xh\n", i, values[0],values[1],values[2],values[3]);
    }
  }

  printf("\tdd\t0FFFFFFFFh\n");

  putchar('\n');
#endif

  puts("Press any key to continue..");
  getch();

  putchar('\n');

  return(0);
}

//*****************************************************************************

/*
CPUIDEmu PROC ; On Entry EAX = Index

        push    esi
        mov     esi,EmuTable

        cmp     eax,0FFFFFFFFh
        jz      cpuid_30

cpuid_10:

        cmp     [esi],eax
        jz      cpuid_20

        add     esi,20

        cmp     dword ptr [esi],0FFFFFFFFh
        jnz     cpuid_10

        jmp     cpuid_30

cpuid_20:

        mov     eax,[esi +  4]
        mov     ebx,[esi +  8]
        mov     ecx,[esi + 12]
        mov     edx,[esi + 16]

cpuid_30:

        pop     esi
        ret

CPUIDEmu ENDP

*/
